* QBXMLTools.prg		(c) 2002, Peisch Custom Software, Inc.
* www.peisch.com
*    Freeware utilites - may be distributed royalty-free.

* 01/26/03 - Added QBXMLCompatibleString function
* 05/10/02 - Original version

* Code is compatible with Visual FoxPro 6 and 7
* Code has been tested with MSXML 4 SP1

* These tools are meant to be generic tools for doing simple queries
* and add procedures for QuickBooks 2002, using QBXML format.

* ASSUMPTIONS:
*    1. You have QuickBooks 2002 installed, are familiar with
*       QBXML and the SDK.
*    2. You have instantiated the QuickBooks COM object before
*       attempting to call any of the methods here, and you have
*       established a session with QuickBooks.
*    3. You have Microsoft's XML DOM parser installed (MSXML 4)
*       (Methods that need to will instantiate the parser.)

* USAGE:
*    (Note: you can copy the code below into a program, select all, and uncomment.)
*!* #INCLUDE ..\INCLUDE\QBXML.H

*!*	* Start by instantiating the MSXML parser:
*!*	   loXMLDoc = CREATEOBJECT( 'MSXML2.DOMDocument' )
*!*	   loXMLDoc.async = .F.

*!*	* Instantiate the QBXML COM object and open a session:
*!*	   loQBRequest = NEWOBJECT('QBXMLRP.RequestProcessor','QBXMLTools.prg')
*!*	   WAIT WINDOW NOWAIT NOCLEAR 'Opening QuickBooks Session. Please wait...'
*!*	   llResult = loQBRequest.OpenConnection('',"My App Name")
*!*	   IF (vartype(llResult) = 'L' and not llResult) or (vartype(llResult) = 'C' and not empty(llResult))
*!*		  * Do some error processing
*!*		   return
*!*	   ENDIF
*!*	   lcTicket = loQBRequest.BeginSession(QBPathAndCompanyFile, QBFILEOPENDONOTCARE)

*!*	* Instantiate the tools and call methods:
*!*	   loQBXML = CREATEOBJECT('QBXMLTOOLS')
*!*	   loQBXML.oQBXML = loQBRequest
*!*	   loQBXML.cTicket = lcTicket

*!*	* A Typical Query of a specific entity:
*!*	   lnResult = loQBXML.QBQuery("Customer", "SomeCustName", "1CUST")
*!*	   DO CASE
*!*	   CASE lnResult = -1		&& There was an error
*!*	   	* Get the QB error code from loQBXML.cErrorCode
*!*	   	* Get the error message from loQBXML.cErrorMessage and do error processing
*!*	   CASE lnResult = 0		&& The customer doesn't exist in QB
*!*	   	* Add the customer, or whatever you want to do
*!*	   OTHERWISE				&& The customer was found in QB
*!*	   	* The ListID of the entity found is in loQBXML.cQBListID
*!*	   	* The entire XML string is in loQBXML.cXMLResponse
*!*	   	* Do some processing
*!*	   ENDCASE

*!*	* A Typical Add:
*!*	   llResult = loQBXML.QBAdd("Customer", "SomeCustName", "1CUST")
*!*	   IF llResult
*!*	   	* The ListID of the entity added is in loQBXML.cQBListID
*!*	   	* The entire XML string is in loQBXML.cXMLResponse
*!*	   	* Do some processing
*!*	   ELSE
*!*	   	* Get the QB error code from loQBXML.cErrorCode
*!*	   	* Get the error message from loQBXML.cErrorMessage and do error processing
*!*	   ENDIF

*!*	* A Typical Query for an entire list:
*!*	   lnResult = loQBXML.QBQuery("Invoice", "", "1INV")
*!*	   DO CASE
*!*	   CASE lnResult = -1		&& There was an error
*!*	   	* Get the QB error code from loQBXML.cErrorCode
*!*	   	* Get the error message from loQBXML.cErrorMessage and do error processing
*!*	   CASE lnResult = 0		&& There are no invoices
*!*	   	* Add the customer, or whatever you want to do
*!*	   OTHERWISE				&& The the list was returned
*!*	   	* The ListID of the entity found is in loQBXML.cQBListID
*!*	   	* The entire XML string is in loQBXML.cXMLResponse
*!*	   	* Do some processing
*!*	   ENDCASE

*!*	* If you are formatting your own XML strings:
*!*	   * QBXMLHeader() will return the standard header required for all QB XML strings
*!*	   * QBXMLFooter() will return the standard footer required for all QB XML strings
*!*    lcXML = loQBXML.QBXMLHeader()
*!*	   * Now add to lcXML
*!*	   lcXML = lcXML+CHR(13)+loQBXML.QBXMLFooter()
*!*	   * or for VFP 7:
*!*	   TEXT TO lcXML TEXTMERGE NOSHOW
*!*	   <<loQBXMLHeader()>>
*!*	   The rest of your XML string gets built up here
*!*	   <<loQBXMLFooter()>>
*!*	   ENDTEXT

*!*	   * Once you've built your own XML strings:
*!*	   * QBSendAdd will send an XML Add request
*!*	   lcQBReturn = loQBXML.QBSendAdd(lcXMLString, lcEntity)
*!*	   * QBSendQuery will send an XML Query request
*!*	   lcQBReturn = loQBXML.QBSendQuery(lcXMLString, lcEntity)
*!*	   * Return values are the same as for the generic add and query methods

*!*	* Other utilities:
*!*	  * To format a date type into form required by QB:
*!*	  lcQBDate = loQBXML.QBDateFormat(ldDate)		&& returns the date as yyyy-mm-dd
*!*   * To convert any string that may have &, <, > or " characters embedded into XML accetable characters:
*!*	  lcStrint = loQBXML.QBCompatibleString(lcOrigString)

DEFINE CLASS QBXMLTools as Session olepublic

oQBXML  = .null.		&& Object reference to QBXML COM object
cTicket = ''			&& The "ticket" we're using to talk to this QBXML session
cErrorMessage = ''		&& Message from last error encountered
cErrorCode = ''			&& Code from last error encountered (QB error codes)
cQBListId = ''			&& The QuickBooks ListID property from the last query or add request
cXMLResponse = ''		&& The XML string returned by QB

*---------------------------------------------------------------------------------
FUNCTION QBQuery
* Checks for existence in QuickBooks for the entity specified.
* If no entity value is passed, all entries are returned and can be found in
* the cXMLReponse property.

* RETURNS: -1 if an error was encountered
*           0 if the entity wasn't found
*           1 if the entity was found

LPARAMETERS tcEntity, tcEntValue, tcReqId
* tcEntity = Type of entity we're checking for, e.g. a customer, an invoice, etc.
*            The entity must be exactly as it appears in the query tag of a QB request
*            (the part of the tag before QueryRq).  TAGS ARE CASE SENSITIVE!
*            For example, the tcEntity for a Customer query is "Customer" because the QB
*            tag is <CustomerQueryRq>. The tcEntity for a customer type query would be 
*            "CustomerType".
*     Current entity queries supported are:
*         Any entities with a NameFilter available
*         Invoice (by RefNumber)
* tcEntValue = Value of the entity we're looking for.  Will use the StartsWith criterion for matching
*              a name filter.  If this parameter isn't passed, the query will be for all records
* tcReqId = The RequestId to pass to the query (per QB documentation)

LOCAL lcQuery, lcQBReturn

this.cErrorCode = ''
this.cErrorMessage = ''

* Validate the parameters
IF VARTYPE(this.oQBXML) <> 'O'
	this.cErrorMessage = 'You must set the oQBXML property in the QBXMLTools object!'
	RETURN -1
ENDIF
IF VARTYPE(this.cTicket) <> 'C'
	this.cErrorMessage = 'You must set the cTicket property in the QBXMLTools object!'
	RETURN -1
ENDIF
IF VARTYPE(tcEntity) <> 'C'
	this.cErrorMessage = 'You must pass the name of the entity to query to the QBQuery method!'
	RETURN -1
ENDIF
IF VARTYPE(tcEntValue) <> 'C' or EMPTY(tcEntValue)
	tcEntValue = ''
ENDIF
IF VARTYPE(tcReqId) <> 'C'
	this.cErrorMessage = 'You must pass the Request ID to the QBQuery method!'
	RETURN -1
ENDIF

lcQuery = this.QBXMLHeader()
lcQuery = lcQuery+CHR(13)+[     <]+tcEntity+[QueryRq requestID = "]+tcReqId+[">]
IF NOT EMPTY(tcEntValue)
	DO CASE
	CASE tcEntity = "Invoice"
		lcQuery = lcQuery+CHR(13)+[        <RefNumber>]+tcEntValue+[</RefNumber>]
	OTHERWISE
		lcQuery = lcQuery+CHR(13)+[        <NameFilter>]
		lcQuery = lcQuery+CHR(13)+[           <MatchCriterion>StartsWith</MatchCriterion>]
		lcQuery = lcQuery+CHR(13)+[           <Name>]+this.QBCompatibleString(tcEntValue)+[</Name>]
		lcQuery = lcQuery+CHR(13)+[        </NameFilter>]
	ENDCASE
ENDIF
lcQuery = lcQuery+CHR(13)+[     </]+tcEntity+[QueryRq>]
lcQuery = lcQuery+CHR(13)+this.QBXMLFooter()

lcQBReturn = this.QBSendQuery(lcQuery, tcEntity)

RETURN lcQBReturn

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBAdd
* Generic add method for QuickBooks.
*  Good for adding entities that only need Name and IsActive fields.
*   (IsActive is assummed to be true)

* RETURNS: .T. if the add was successful, otherwise .F.

LPARAMETERS tcEntity, tcEntValue, tcRequestId
* tcEntity = Type of entity we're adding, e.g. a class, terms, etc.
*            The entity must be exactly as it appears in the add tag of a QB request
*            (the part of the tag before AddRq).  TAGS ARE CASE SENSITIVE!
*            For example, the tcEntity for a Class query is "Class" because the QB
*            tag is <ClassAddRq>. The tcEntity for terms would be "Terms".
* tcEntValue = Value to fill in for entity name
* tcRequestId = Recno to use in RequestId

LOCAL lcEntityAdd, lcQBReturn

this.cErrorCode = ''
this.cErrorMessage = ''

* Validate the parameters
IF VARTYPE(this.oQBXML) <> 'O'
	this.cErrorMessage = 'You must set the oQBXML property in the QBXMLTools object!'
	RETURN .F.
ENDIF
IF VARTYPE(this.cTicket) <> 'C'
	this.cErrorMessage = 'You must set the cTicket property in the QBXMLTools object!'
	RETURN .F.
ENDIF
IF VARTYPE(tcEntity) <> 'C'
	this.cErrorMessage = 'You must pass the name of the entity to query to the QBAdd method!'
	RETURN .F.
ENDIF
IF VARTYPE(tcEntValue) <> 'C'
	this.cErrorMessage = 'You must pass the value of the entity to query the QBAdd method!'
	RETURN .F.
ENDIF
IF VARTYPE(tcRequestId) <> 'C'
	this.cErrorMessage = 'You must pass the Request ID to the QBQuery method!'
	RETURN .F.
ENDIF

lcEntityAdd = this.QBXMLHeader()
lcEntityAdd = lcEntityAdd+CHR(13)+[        <]+tcEntity+[AddRq requestID="]+tcRequestID+[">]
lcEntityAdd = lcEntityAdd+CHR(13)+[           <]+tcEntity+[Add>]
lcEntityAdd = lcEntityAdd+CHR(13)+[              <Name>]+this.QBCompatibleString(tcEntValue)+[</Name>]
lcEntityAdd = lcEntityAdd+CHR(13)+[              <IsActive>true</IsActive>]
lcEntityAdd = lcEntityAdd+CHR(13)+[           </]+tcEntity+[Add>]
lcEntityAdd = lcEntityAdd+CHR(13)+[        </]+tcEntity+[AddRq>]
lcEntityAdd = lcEntityAdd+CHR(13)+this.QBXMLFooter()

lcQBReturn = this.QBSendAdd(lcEntityAdd, tcEntity)

RETURN lcQBReturn

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBXMLFooter
*  Returns standard closing tags for QBXML strings
LOCAL lcOutput

lcOutput = [      </QBXMLMsgsRq>]
lcOutput = lcOutput+chr(13)+[   </QBXML>]

RETURN lcOutput

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBXMLHeader
*  Returns standard opening tags for QBXML strings
LOCAL lcOutput

lcOutput = [<?xml version="1.0" ?>]
lcOutput = lcOutput+chr(13)+[<!DOCTYPE QBXML PUBLIC  "-//INTUIT//DTD QBXML QBD 1.1//EN" "http://developer.intuit.com">]
lcOutput = lcOutput+chr(13)+[   <QBXML>]
lcOutput = lcOutput+chr(13)+[      <QBXMLMsgsRq onError = "continueOnError">]

RETURN lcOutput

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBSendQuery
* Generic method to send a query request to QuickBooks

* RETURNS: -1 if an error was encountered
*           0 if the entity wasn't found
*           1 if the entity was found

LPARAMETERS tcXMLMsg, tcEntity
* tcXMLMsg = XML string to send to QB
* tcEntity = Name of the entity we're adding
LOCAL lcQBReturn, loXMLDoc, llRet, loNodeList, loAttRSNamedNodeMap, lcRetStatusCode, lcRetStatusSeverity, ;
      lcRetStatusMessage

this.cErrorCode = ''
this.cErrorMessage = ''

IF VARTYPE(this.oQBXML) <> 'O'
	this.cErrorMessage = 'You must set the oQBXML property in the QBXMLTools object!'
	RETURN .F.
ENDIF
IF VARTYPE(this.cTicket) <> 'C'
	this.cErrorMessage = 'You must set the cTicket property in the QBXMLTools object!'
	RETURN .F.
endif
IF VARTYPE(tcXMLMsg) <> 'C'
	this.cErrorMessage = 'You must pass the XML string to the QBSendQuery method!'
	RETURN .F.
ENDIF
IF VARTYPE(tcEntity) <> 'C'
	this.cErrorMessage = 'You must pass the name of the entity you are sending to the QBSendQuery method!'
	RETURN .F.
ENDIF

*strtofile(tcxmlmsg,"temp.xml")		&& Use for debugging
lcQBReturn = this.oQBXML.ProcessRequest(this.cTicket, tcXMLMsg)
this.cXMLResponse = lcQBReturn

IF EMPTY(lcQBReturn)
	this.cErrorMessage = "QuickBooks returned an empty string after processing request!"
	RETURN -1
ENDIF

loXMLDoc = CREATEOBJECT( 'MSXML2.DOMDocument' )
loXMLDoc.async = .F.

llRet = loXMLDoc.loadXML(lcQBReturn)
IF NOT llRet
	this.cErrorMessage = loXMLDoc.parseError.reason
	RETURN -1
ENDIF

loNodeList = loXMLDoc.getElementsByTagName(tcEntity+"QueryRs")
loAttRSNamedNodeMap = loNodeList.item(0).attributes
lcRetStatusCode = loAttRSNamedNodeMap.getNamedItem( "statusCode" ).nodeValue
lcRetStatusSeverity = loAttRSNamedNodeMap.getNamedItem( "statusSeverity" ).nodeValue
lcRetStatusMessage = loAttRSNamedNodeMap.getNamedItem( "statusMessage" ).nodeValue
IF lcRetStatusCode <> "0"
	IF inlist(lcRetStatusCode,"1","500")
		* Entity wasn't found
		this.cQBListId = ''
		RETURN 0
	ELSE
		* Some other error
		this.cQBListId = ''
		this.cErrorCode = lcRetStatusCode
		this.cErrorMessage = lcRetStatusMessage
		RETURN -1
	ENDIF
else
	* Entity was found
*	lcStatusId = loNodeList.getValueFromQName("ListID")
* Replace the following code with the correct XML DOM code once I figure it out
lloc = at('<ListID>',lcQBReturn)
lloc2 = at('</ListID>',lcQBReturn)
lcStatusId = substr(lcQBReturn,lloc+8,lloc2-(lloc+8))
	this.cQBListId = lcStatusId
	RETURN 1
ENDIF

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBSendAdd
* Generic method to send an add request to QuickBooks

* Returns .T. if add was successful, otherwise .F.

LPARAMETERS tcXMLMsg, tcEntity
* tcXMLMsg = XML string to send to QB
* tcEntity = Name of the entity we're adding
LOCAL lcQBReturn, loXMLDoc, llRet, loNodeList, loAttRSNamedNodeMap, lcRetStatusCode, lcRetStatusSeverity, ;
      lcRetStatusMessage

this.cErrorCode = ''
this.cErrorMessage = ''

IF VARTYPE(this.oQBXML) <> 'O'
	this.cErrorMessage = 'You must set the oQBXML property in the QBXMLTools object!'
	RETURN .F.
ENDIF
IF VARTYPE(this.cTicket) <> 'C'
	this.cErrorMessage = 'You must set the cTicket property in the QBXMLTools object!'
	RETURN .F.
endif
IF VARTYPE(tcXMLMsg) <> 'C'
	this.cErrorMessage = 'You must pass the XML string to the QBSendAdd method!'
	RETURN .F.
ENDIF
IF VARTYPE(tcEntity) <> 'C'
	this.cErrorMessage = 'You must pass the name of the entity you are sending to the QBSendAdd method!'
	RETURN .F.
ENDIF

*strtofile(tcxmlmsg,"temp.xml")		&& Use for debugging
lcQBReturn = this.oQBXML.ProcessRequest(this.cTicket, tcXMLMsg)
this.cXMLResponse = lcQBReturn

IF EMPTY(lcQBReturn)
	RETURN .F.
ENDIF

loXMLDoc = CREATEOBJECT( 'MSXML2.DOMDocument' )
loXMLDoc.async = .F.

llRet = loXMLDoc.loadXML(lcQBReturn)
IF not llRet
	this.cErrorMessage = loXMLDoc.parseError.reason
	RETURN .F.
ENDIF

loNodeList = loXMLDoc.getElementsByTagName(tcEntity+"AddRs")
loAttRSNamedNodeMap = loNodeList.item(0).attributes
lcRetStatusCode = loAttRSNamedNodeMap.getNamedItem( "statusCode" ).nodeValue
lcRetStatusSeverity = loAttRSNamedNodeMap.getNamedItem( "statusSeverity" ).nodeValue
lcRetStatusMessage = loAttRSNamedNodeMap.getNamedItem( "statusMessage" ).nodeValue
IF lcRetStatusCode <> "0"
	this.cErrorCode = lcRetStatusCode
	this.cErrorMessage = lcRetStatusMessage
	RETURN .F.
ELSE
	* Success!
*	lcStatusId = loNodeList.getValueFromQName("ListID")
* Replace the following code with the correct XML DOM code once I figure it out
	lloc = at('<ListID>',lcQBReturn)
	lloc2 = at('</ListID>',lcQBReturn)
	lcStatusId = substr(lcQBReturn,lloc+8,lloc2-(lloc+8))
	this.cQBListId = lcStatusId
	RETURN .T.
ENDIF

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBDateFormat
* Formats a date into the format needed for QuickBooks XML, i.e. yyyy-mm-dd.
LPARAMETERS tdDate

LOCAL lcOldCentury, lcOldDate, lcReturn

IF VARTYPE(tdDate) <> 'D'
	this.cErrorMessage = 'No date passed to the QBDateFormat method!'
	RETURN ''
ENDIF

lcOldCentury = SET('Century')
lcOldDate = SET('Date')
SET CENTURY ON
SET date TO ANSI

lcReturn = CHRTRAN(DTOC(tdDate),'.','-')

SET date &lcOldDate
SET CENTURY &lcOldCentury

RETURN lcReturn

ENDFUNC
*-----------------------------------------------------------------------------------
FUNCTION QBXMLCompatibleString
* Converts a string to a QuickBooks compatible string for Name fields
lparameter tcString
local lcString

lcString = alltrim(tcString)
lcString = strtran(lcString,"&","&amp;")
lcString = strtran(lcString,"<","&lt;")
lcString = strtran(lcString,">","&gt;")
lcString = strtran(lcString,'"',"&quot;")

return lcString
*-----------------------------------------------------------------------------------

ENDDEFINE
************************************************************************************